import platform
from datetime import datetime, timezone

import pytest
from pyspark.sql.functions import lit
from pyspark.sql.types import Row

from pysail.tests.spark.utils import strict


@pytest.mark.parametrize(
    ("session_timezone", "local_timezone", "data"),
    [
        (
            "America/Los_Angeles",
            "Asia/Shanghai",
            [
                (datetime(1970, 1, 1), "1969-12-31 08:00:00", datetime(1970, 1, 1)),  # noqa: DTZ001
                (datetime(1970, 1, 1, tzinfo=timezone.utc), "1969-12-31 16:00:00", datetime(1970, 1, 1, 8)),  # noqa: DTZ001
                (datetime(2024, 11, 3, 16, 30), "2024-11-03 01:30:00", datetime(2024, 11, 3, 16, 30)),  # noqa: DTZ001
                (datetime(2024, 11, 3, 17, 30), "2024-11-03 01:30:00", datetime(2024, 11, 3, 17, 30)),  # noqa: DTZ001
                (datetime(2025, 3, 9, 17, 30), "2025-03-09 01:30:00", datetime(2025, 3, 9, 17, 30)),  # noqa: DTZ001
                (datetime(2025, 3, 9, 18, 30), "2025-03-09 03:30:00", datetime(2025, 3, 9, 18, 30)),  # noqa: DTZ001
            ],
        ),
        (
            "America/Los_Angeles",
            "America/Los_Angeles",
            [
                (datetime(1970, 1, 1), "1970-01-01 00:00:00", datetime(1970, 1, 1)),  # noqa: DTZ001
                (datetime(1970, 1, 1, tzinfo=timezone.utc), "1969-12-31 16:00:00", datetime(1969, 12, 31, 16)),  # noqa: DTZ001
                # The PySpark client does not seem to respect the `fold` parameter for local `datetime` objects.
                # This is a bug on the client side, and no special handling is needed for the server.
                (datetime(2024, 11, 3, 1, 30), "2024-11-03 01:30:00", datetime(2024, 11, 3, 1, 30, fold=1)),  # noqa: DTZ001
                (datetime(2024, 11, 3, 1, 30, fold=0), "2024-11-03 01:30:00", datetime(2024, 11, 3, 1, 30, fold=1)),  # noqa: DTZ001
                (datetime(2024, 11, 3, 1, 30, fold=1), "2024-11-03 01:30:00", datetime(2024, 11, 3, 1, 30, fold=1)),  # noqa: DTZ001
                (datetime(2025, 3, 9, 2, 30), "2025-03-09 03:30:00", datetime(2025, 3, 9, 3, 30)),  # noqa: DTZ001
                (datetime(2025, 3, 9, 3, 30), "2025-03-09 03:30:00", datetime(2025, 3, 9, 3, 30)),  # noqa: DTZ001
            ],
        ),
    ],
    indirect=["session_timezone", "local_timezone"],
)
@pytest.mark.skipif(platform.system() == "Windows", reason="`time.tzset()` is not available on Windows")
def test_datetime_literal(spark, session_timezone, local_timezone, data):  # noqa: ARG001
    for dt, k, v in data:
        assert spark.range(1).select(lit(dt)).collect() == [strict(Row(**{f"TIMESTAMP '{k}'": v}))]



import pandas as pd
from pandas.testing import assert_frame_equal


def test_dataframe_sample_replazement_seed(spark):
    df = spark.createDataFrame([(0), (1), (2), (3), (4), (5), (6), (7), (8), (9)], ["id"])
    df2 = df.sample(True, 0.5, 1)

    assert_frame_equal(
        df2.toPandas(),
        pd.DataFrame({"id": [0, 0, 3, 4, 8]}),
    )



import pandas as pd
import pyspark.sql.functions as F  # noqa: N812
import pytest
from pandas.testing import assert_frame_equal

from pysail.tests.spark.utils import escape_sql_string_literal, is_jvm_spark


@pytest.fixture(autouse=True)
def tables(spark, tmp_path):
    location = str(tmp_path / "t1")
    spark.sql(f"CREATE TABLE t1 (id LONG, name STRING, age LONG) LOCATION '{escape_sql_string_literal(location)}'")
    yield
    spark.sql("DROP TABLE t1")
    spark.sql("DROP TABLE IF EXISTS t2")
    spark.sql("DROP TABLE IF EXISTS t3")
    spark.sql("DROP TABLE IF EXISTS t4")
    spark.sql("DROP TABLE IF EXISTS t5")


def test_insert_into_with_values(spark):
    spark.sql("INSERT INTO t1 VALUES (101, 'Alice', 20)")
    df = spark.createDataFrame([(102, "Bob", 30)])
    df.write.insertInto("t1")

    actual = spark.sql("SELECT * FROM t1 WHERE id IN (101, 102) ORDER BY id").toPandas()
    expected = pd.DataFrame(
        {"id": [101, 102], "name": ["Alice", "Bob"], "age": [20, 30]},
    )
    assert_frame_equal(actual, expected)


def test_insert_into_by_position(spark):
    # The field names are ignored.
    spark.sql("INSERT INTO t1 SELECT 201 AS age, 'Alice' AS name, 21 AS id")
    df = spark.createDataFrame([(202, "Bob", 31)], schema=["age", "name", "id"])
    df.write.insertInto("t1")

    actual = spark.sql("SELECT * FROM t1 WHERE id IN (201, 202) ORDER BY id").toPandas()
    expected = pd.DataFrame(
        {"id": [201, 202], "name": ["Alice", "Bob"], "age": [21, 31]},
    )
    assert_frame_equal(actual, expected)


def test_insert_into_by_name(spark):
    spark.sql("INSERT INTO t1 (age, name, id) VALUES (21, 'Alice', 301)")
    spark.sql("INSERT INTO t1 BY NAME SELECT 31 AS age, 'Bob' AS name, 302 AS id")

    actual = spark.sql("SELECT * FROM t1 WHERE id IN (301, 302) ORDER BY id").toPandas()
    expected = pd.DataFrame(
        {"id": [301, 302], "name": ["Alice", "Bob"], "age": [21, 31]},
    )
    assert_frame_equal(actual, expected)


def test_insert_into_with_invalid_options(spark):
    df = spark.createDataFrame([(101, "Alice", 20)])

    with pytest.raises(Exception, match="partition"):
        df.write.partitionBy("id").insertInto("t1")


@pytest.mark.skipif(not is_jvm_spark(), reason="`INSERT OVERWRITE` is not supported in Sail yet")
def test_insert_overwrite(spark):
    spark.sql("INSERT INTO t1 VALUES (401, 'Alice', 22)")
    spark.sql("INSERT OVERWRITE t1 VALUES (402, 'Bob', 32)")
    actual = spark.sql("SELECT * FROM t1").toPandas()
    expected = pd.DataFrame(
        {"id": [402], "name": ["Bob"], "age": [32]},
    )
    assert_frame_equal(actual, expected)

    df = spark.createDataFrame([(403, "Alice", 25)])
    df.write.insertInto("t1", overwrite=True)
    actual = spark.sql("SELECT * FROM t1").toPandas()
    expected = pd.DataFrame(
        {"id": [403], "name": ["Alice"], "age": [25]},
    )
    assert_frame_equal(actual, expected)

    df = spark.createDataFrame([(404, "Bob", 35)])
    df.write.mode("overwrite").insertInto("t1")
    actual = spark.sql("SELECT * FROM t1").toPandas()
    expected = pd.DataFrame(
        {"id": [404], "name": ["Bob"], "age": [35]},
    )
    assert_frame_equal(actual, expected)


def test_save_as_table(spark, tmp_path):
    location = str(tmp_path / "t2")
    df = spark.createDataFrame([(1001, "Alice")], schema="id LONG, name STRING")

    def expected(n: int):
        return pd.DataFrame(
            {"id": [1001] * n, "name": ["Alice"] * n},
        )

    df.write.saveAsTable("t2", path=location)
    actual = spark.sql("SELECT * FROM t2").toPandas()
    assert_frame_equal(actual, expected(1))

    if not is_jvm_spark():
        # The "ignore" mode seems broken in Spark Connect.
        df.write.saveAsTable("t2", mode="ignore", path=location)
        actual = spark.sql("SELECT * FROM t2").toPandas()
        assert_frame_equal(actual, expected(1))

    with pytest.raises(Exception, match=".*"):
        df.write.saveAsTable("t2", mode="error", location=location)

    with pytest.raises(Exception, match=".*"):
        df.write.saveAsTable("t2", location=location)

    df.write.saveAsTable("t2", mode="append")
    actual = spark.sql("SELECT * FROM t2").toPandas()
    assert_frame_equal(actual, expected(2))

    # `saveAsTable` matches columns by name.
    df.select("name", "id").write.saveAsTable("t2", mode="append")
    actual = spark.sql("SELECT * FROM t2").toPandas()
    assert_frame_equal(actual, expected(3))

    with pytest.raises(Exception, match=".*"):
        df.select(F.col("name").alias("n"), "id").write.saveAsTable("t2", mode="append")

    with pytest.raises(Exception, match=".*"):
        df.select("id").write.saveAsTable("t2", mode="append")

    if is_jvm_spark():
        # The "overwrite" mode is not supported in Sail yet.
        df.write.saveAsTable("t2", mode="overwrite", path=location)
        actual = spark.sql("SELECT * FROM t2").toPandas()
        assert_frame_equal(actual, expected(1))


@pytest.mark.skipif(is_jvm_spark(), reason="Spark does not handle v1 and v2 tables properly")
def test_write_to(spark, tmp_path):
    location = str(tmp_path / "t3")
    df = spark.createDataFrame([(2002, "Bob")], schema="id LONG, name STRING")

    def expected(n: int):
        return pd.DataFrame(
            {"id": [2002] * n, "name": ["Bob"] * n},
        )

    with pytest.raises(Exception, match=".*"):
        # The table does not exist yet.
        df.writeTo("t3").option("location", location).replace()

    df.writeTo("t3").option("path", location).create()
    actual = spark.sql("SELECT * FROM t3").toPandas()
    assert_frame_equal(actual, expected(1))

    df.writeTo("t3").append()
    actual = spark.sql("SELECT * FROM t3").toPandas()
    assert_frame_equal(actual, expected(2))

    pytest.skip("replace and overwrite are not supported in Sail yet")

    df.writeTo("t3").option("location", location).replace()
    actual = spark.sql("SELECT * FROM t3").toPandas()
    assert_frame_equal(actual, expected(1))

    df.writeTo("t3").option("path", location).createOrReplace()
    actual = spark.sql("SELECT * FROM t3").toPandas()
    assert_frame_equal(actual, expected(1))

    df.writeTo("t3").option("location", location).overwrite(F.lit(True))
    actual = spark.sql("SELECT * FROM t3").toPandas()
    assert_frame_equal(actual, expected(1))


@pytest.mark.skipif(not is_jvm_spark(), reason="the options overwrite logic is not fully compatible yet")
def test_write_options(spark, tmp_path):
    location = str(tmp_path / "t4")
    spark.sql(f"""
        CREATE TABLE t4 (a STRING, b STRING) USING CSV
        LOCATION '{escape_sql_string_literal(location)}'
        OPTIONS (header 'false', delimiter '|')
    """)

    df = spark.createDataFrame([("1-2", "3")], schema="a STRING, b STRING")
    # Spark requires `.format()` to match the table definition.
    # If the format is not specified, the default format is used, and an exception is raised
    # if it does not match the table definition.
    df.write.format("csv").saveAsTable("t4", mode="append")

    df = spark.createDataFrame([("1|2", "3")], schema="a STRING, b STRING")
    df.write.format("csv").option("delimiter", "-").saveAsTable("t4", mode="append")
    df.write.format("csv").option("SEP", "-").saveAsTable("t4", mode="append")

    # Spark rewrites the entire table with the new options.
    assert sorted(spark.read.format("csv").table("t4").collect()) == [
        ("1-2", "3"),
        ("1|2", "3"),
        ("1|2", "3"),
    ]
    # Spark ignores the options if it conflicts with the table definition.
    assert sorted(spark.read.format("csv").option("delimiter", "|").table("t4").collect()) == [
        ("1-2", "3"),
        ("1|2", "3"),
        ("1|2", "3"),
    ]


def test_write_with_partition_columns(spark, tmp_path):
    location = str(tmp_path / "t5")
    spark.sql(f"""
        CREATE TABLE t5 (a INT, b STRING)
        USING PARQUET
        LOCATION '{escape_sql_string_literal(location)}'
        PARTITIONED BY (a)
    """)
    df = spark.createDataFrame([(42, "foo")], schema="a INT, b STRING")
    df.write.partitionBy("a").format("parquet").saveAsTable("t5", mode="append")

    with pytest.raises(Exception, match=".*"):
        df.write.partitionBy("b").format("parquet").saveAsTable("t5", mode="append")

    assert spark.read.format("parquet").table("t5").select("a", "b").collect() == [
        (42, "foo"),
    ]

    if is_jvm_spark():
        pytest.skip("Spark cannot infer partition columns when reading paths")

    assert spark.read.parquet(location).select("a", "b").collect() == [
        ("42", "foo"),
    ]



import datetime

import pyarrow as pa
import pytest
from pyiceberg.partitioning import (
    BucketTransform,
    DayTransform,
    HourTransform,
    IdentityTransform,
    MonthTransform,
    PartitionField,
    PartitionSpec,
    TruncateTransform,
    YearTransform,
)
from pyiceberg.schema import Schema
from pyiceberg.types import DateType, IntegerType, NestedField, StringType, TimestampType

from .utils import create_sql_catalog  # noqa: TID252


def _make_common_schema():
    return Schema(
        NestedField(1, "number", IntegerType(), required=True),
        NestedField(2, "letter", StringType(), required=False),
        NestedField(3, "ts", TimestampType(), required=False),
        NestedField(4, "dt", DateType(), required=False),
    )


def _append_sample_data(table):
    start_date = datetime.date(2023, 3, 1)
    start_dt = datetime.datetime(2023, 3, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)
    letters = [
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
    ]
    numbers = list(range(1, 13))
    dts = [start_date + datetime.timedelta(days=i) for i in range(12)]
    tss = [start_dt + datetime.timedelta(days=i) for i in range(12)]

    arrow_schema = pa.schema(
        [
            pa.field("number", pa.int32(), nullable=False),
            pa.field("letter", pa.string(), nullable=True),
            pa.field("ts", pa.timestamp("us"), nullable=True),
            pa.field("dt", pa.date32(), nullable=True),
        ]
    )

    for i in range(0, 12, 4):
        arrays = [
            pa.array(numbers[i : i + 4], type=pa.int32()),
            pa.array(letters[i : i + 4], type=pa.string()),
            pa.array(tss[i : i + 4], type=pa.timestamp("us")),
            pa.array(dts[i : i + 4], type=pa.date32()),
        ]
        batch = pa.Table.from_arrays(arrays, schema=arrow_schema)
        table.append(batch)


@pytest.mark.parametrize(
    ("table_name", "spec", "predicate_column", "predicate_value", "expected_numbers"),
    [
        pytest.param(
            # FIXME: this test is failing in `local-cluster` mode because of
            # the proto issue related to `PartitionedFile`.
            "default.test_partitioned_by_identity",
            PartitionSpec(PartitionField(3, 1001, IdentityTransform(), "ts")),
            "ts",
            "'2023-03-05T00:00:00+00:00'",
            {5, 6, 7, 8, 9, 10, 11, 12},
            id="partition-by-identity",
        ),
        pytest.param(
            "default.test_partitioned_by_years",
            PartitionSpec(PartitionField(4, 1002, YearTransform(), "dt_year")),
            "dt",
            "'2023-03-05'",
            {5, 6, 7, 8, 9, 10, 11, 12},
            id="partition-by-years",
        ),
        pytest.param(
            "default.test_partitioned_by_months",
            PartitionSpec(PartitionField(4, 1003, MonthTransform(), "dt_month")),
            "dt",
            "'2023-03-05'",
            {5, 6, 7, 8, 9, 10, 11, 12},
            id="partition-by-months",
        ),
        pytest.param(
            "default.test_partitioned_by_days",
            PartitionSpec(PartitionField(3, 1004, DayTransform(), "ts_day")),
            "ts",
            "'2023-03-05T00:00:00+00:00'",
            {5, 6, 7, 8, 9, 10, 11, 12},
            id="partition-by-days",
        ),
        pytest.param(
            "default.test_partitioned_by_hours",
            PartitionSpec(PartitionField(3, 1005, HourTransform(), "ts_hour")),
            "ts",
            "'2023-03-05T00:00:00+00:00'",
            {5, 6, 7, 8, 9, 10, 11, 12},
            id="partition-by-hours",
        ),
        pytest.param(
            "default.test_partitioned_by_truncate",
            PartitionSpec(PartitionField(2, 1006, TruncateTransform(1), "letter_trunc")),
            "letter",
            "'e'",
            {5, 6, 7, 8, 9, 10, 11, 12},
            id="partition-by-truncate",
        ),
        pytest.param(
            "default.test_partitioned_by_bucket",
            PartitionSpec(PartitionField(1, 1007, BucketTransform(8), "number_bucket")),
            "number",
            "5",
            {5, 6, 7, 8, 9, 10, 11, 12},
            id="partition-by-bucket",
        ),
    ],
)
def test_partition_transform_pruning(
    spark, tmp_path, table_name, spec, predicate_column, predicate_value, expected_numbers
):
    catalog = create_sql_catalog(tmp_path)
    schema = _make_common_schema()
    table = catalog.create_table(identifier=table_name, schema=schema, partition_spec=spec)
    try:
        _append_sample_data(table)
        path = table.location()

        predicate = f"{predicate_column} >= {predicate_value}"
        df = spark.read.format("iceberg").load(path).filter(predicate).select("number")
        result = {r[0] for r in df.collect()}
        assert result == expected_numbers
    finally:
        catalog.drop_table(table_name)



import datetime as dt
import platform
from typing import List, Tuple  # noqa: UP035

import pytest
from pyiceberg.partitioning import (
    BucketTransform,
    DayTransform,
    HourTransform,
    IdentityTransform,
    MonthTransform,
    PartitionField,
    PartitionSpec,
    TruncateTransform,
    YearTransform,
)
from pyiceberg.schema import Schema
from pyiceberg.types import DateType, IntegerType, NestedField, StringType, TimestampType

from pysail.tests.spark.utils import escape_sql_string_literal

from .utils import create_sql_catalog, pyiceberg_to_pandas  # noqa: TID252


def _common_schema() -> Schema:
    return Schema(
        NestedField(1, "number", IntegerType(), required=True),
        NestedField(2, "letter", StringType(), required=False),
        NestedField(3, "ts", TimestampType(), required=False),
        NestedField(4, "dt", DateType(), required=False),
    )


def _build_sample_rows() -> List[Tuple[int, str, dt.datetime, dt.date]]:  # noqa: UP006
    start_date = dt.date(2023, 3, 1)
    start_ts = dt.datetime(2023, 3, 1, 0, 0, 0, tzinfo=dt.timezone.utc)
    letters = [
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
    ]
    rows: List[Tuple[int, str, dt.datetime, dt.date]] = []  # noqa: UP006
    for i in range(12):
        rows.append((i + 1, letters[i], start_ts + dt.timedelta(days=i), start_date + dt.timedelta(days=i)))  # noqa: PERF401
    return rows


@pytest.mark.parametrize(
    ("table_name", "spec", "predicate_column", "predicate_value", "expected_numbers"),
    [
        (
            "default.write_partition_identity",
            PartitionSpec(PartitionField(3, 1001, IdentityTransform(), "ts")),
            "ts",
            "'2023-03-05T00:00:00+00:00'",
            {5, 6, 7, 8, 9, 10, 11, 12},
        ),
        (
            "default.write_partition_year",
            PartitionSpec(PartitionField(4, 1002, YearTransform(), "dt_year")),
            "dt",
            "'2023-03-05'",
            {5, 6, 7, 8, 9, 10, 11, 12},
        ),
        (
            "default.write_partition_month",
            PartitionSpec(PartitionField(4, 1003, MonthTransform(), "dt_month")),
            "dt",
            "'2023-03-05'",
            {5, 6, 7, 8, 9, 10, 11, 12},
        ),
        (
            "default.write_partition_day",
            PartitionSpec(PartitionField(3, 1004, DayTransform(), "ts_day")),
            "ts",
            "'2023-03-05T00:00:00+00:00'",
            {5, 6, 7, 8, 9, 10, 11, 12},
        ),
        (
            "default.write_partition_hour",
            PartitionSpec(PartitionField(3, 1005, HourTransform(), "ts_hour")),
            "ts",
            "'2023-03-05T00:00:00+00:00'",
            {5, 6, 7, 8, 9, 10, 11, 12},
        ),
        (
            "default.write_partition_trunc",
            PartitionSpec(PartitionField(2, 1006, TruncateTransform(1), "letter_trunc")),
            "letter",
            "'e'",
            {5, 6, 7, 8, 9, 10, 11, 12},
        ),
        (
            "default.write_partition_bucket",
            PartitionSpec(PartitionField(1, 1007, BucketTransform(8), "number_bucket")),
            "number",
            "5",
            {5, 6, 7, 8, 9, 10, 11, 12},
        ),
    ],
)
@pytest.mark.skipif(platform.system() == "Windows", reason="may not work on Windows")
def test_partitioned_write_then_sail_read(
    spark, tmp_path, table_name, spec, predicate_column, predicate_value, expected_numbers
):
    catalog = create_sql_catalog(tmp_path)
    schema = _common_schema()
    table = catalog.create_table(identifier=table_name, schema=schema, partition_spec=spec)
    try:
        rows = _build_sample_rows()
        df = spark.createDataFrame(rows, schema="number INT, letter STRING, ts TIMESTAMP, dt DATE")
        df.write.format("iceberg").mode("overwrite").save(table.location())

        predicate = f"{predicate_column} >= {predicate_value}"
        out_df = spark.read.format("iceberg").load(table.location()).filter(predicate).select("number")
        result = {r[0] for r in out_df.collect()}
        assert result == expected_numbers
    finally:
        catalog.drop_table(table_name)


@pytest.mark.parametrize(
    ("table_name", "spec"),
    [
        (
            "default.write_partition_identity_all",
            PartitionSpec(PartitionField(3, 1001, IdentityTransform(), "ts")),
        ),
        (
            "default.write_partition_year_all",
            PartitionSpec(PartitionField(4, 1002, YearTransform(), "dt_year")),
        ),
        (
            "default.write_partition_month_all",
            PartitionSpec(PartitionField(4, 1003, MonthTransform(), "dt_month")),
        ),
        (
            "default.write_partition_day_all",
            PartitionSpec(PartitionField(3, 1004, DayTransform(), "ts_day")),
        ),
        (
            "default.write_partition_hour_all",
            PartitionSpec(PartitionField(3, 1005, HourTransform(), "ts_hour")),
        ),
        (
            "default.write_partition_trunc_all",
            PartitionSpec(PartitionField(2, 1006, TruncateTransform(1), "letter_trunc")),
        ),
        (
            "default.write_partition_bucket_all",
            PartitionSpec(PartitionField(1, 1007, BucketTransform(8), "number_bucket")),
        ),
    ],
)
@pytest.mark.skipif(platform.system() == "Windows", reason="may not work on Windows")
def test_partitioned_write_then_pyiceberg_read_all(spark, tmp_path, table_name, spec):
    catalog = create_sql_catalog(tmp_path)
    schema = _common_schema()
    table = catalog.create_table(identifier=table_name, schema=schema, partition_spec=spec)
    try:
        rows = _build_sample_rows()
        df = spark.createDataFrame(rows, schema="number INT, letter STRING, ts TIMESTAMP, dt DATE")
        df.write.format("iceberg").mode("overwrite").save(table.location())

        py_tbl = catalog.load_table(table_name)
        actual = pyiceberg_to_pandas(py_tbl, sort_by="number")

        expected_letters = [
            "a",
            "b",
            "c",
            "d",
            "e",
            "f",
            "g",
            "h",
            "i",
            "j",
            "k",
            "l",
        ]
        assert len(actual) == 12  # noqa: PLR2004
        assert list(actual["number"]) == list(range(1, 13))
        assert list(actual["letter"]) == expected_letters
    finally:
        catalog.drop_table(table_name)


@pytest.mark.skipif(platform.system() == "Windows", reason="may not work on Windows")
def test_iceberg_partition_writes_sql(spark, tmp_path):
    warehouse = tmp_path / "warehouse"
    warehouse.mkdir(parents=True, exist_ok=True)
    path_single = f"file://{warehouse / 'ice_single'}"
    spark.sql(
        f"""
        CREATE TABLE tmp_ice_single (
            id INT,
            event STRING,
            score INT
        ) USING iceberg
        PARTITIONED BY (id)
        LOCATION '{escape_sql_string_literal(path_single)}'
        """
    )
    try:
        df_single = spark.createDataFrame(
            [(10, "A", 1), (11, "B", 2), (12, "A", 3)], schema="id INT, event STRING, score INT"
        )
        df_single.write.format("iceberg").mode("append").save(path_single)
        out_single = spark.read.format("iceberg").load(path_single).sort("id")
        assert [r[0] for r in out_single.select("id").collect()] == [10, 11, 12]
    finally:
        spark.sql("DROP TABLE IF EXISTS tmp_ice_single")

    path_multi = f"file://{warehouse / 'ice_multi'}"
    spark.sql(
        f"""
        CREATE TABLE tmp_ice_multi (
            id INT,
            region INT,
            category INT,
            value INT
        ) USING iceberg
        PARTITIONED BY (region, category)
        LOCATION '{escape_sql_string_literal(path_multi)}'
        """
    )
    try:
        df_multi = spark.createDataFrame(
            [(1, 1, 1, 100), (2, 1, 2, 200), (3, 2, 1, 300), (4, 2, 2, 400)],
            schema="id INT, region INT, category INT, value INT",
        )
        df_multi.write.format("iceberg").mode("append").save(path_multi)
        parts = spark.read.format("iceberg").load(path_multi).select("region", "category").distinct()
        actual = {f"region={r[0]}/category={r[1]}" for r in parts.collect()}
        assert actual == {"region=1/category=1", "region=1/category=2", "region=2/category=1", "region=2/category=2"}
    finally:
        spark.sql("DROP TABLE IF EXISTS tmp_ice_multi")



import pandas as pd
import pyspark.sql.functions as F  # noqa: N812


def partition_count(df):
    def counter(_iterator):
        yield pd.DataFrame({"n": [1]})

    return df.mapInPandas(counter, schema="n: long").count()


def test_explicit_repartition(spark):
    assert partition_count(spark.range(0, 10, 1, 2)) == 2  # noqa: PLR2004
    assert partition_count(spark.range(0, 10, 1, 3).select("id", F.lit("foo").alias("a")).repartition("a")) == 3  # noqa: PLR2004
    assert partition_count(spark.sql("SELECT 1 AS a, 'foo' as b").repartition(5)) == 5  # noqa: PLR2004
    assert partition_count(spark.sql("SELECT 1 AS a, 'foo' as b").repartition(6, "b", "a")) == 6  # noqa: PLR2004
    assert partition_count(spark.sql("SELECT 1 AS a, 'foo' as b").repartition("a")) == 1


def test_explicit_coalesce(spark):
    assert partition_count(spark.range(0, 10, 1, 2).coalesce(1)) == 1
    assert partition_count(spark.range(0, 10, 1, 2).coalesce(3)) == 3  # noqa: PLR2004
    assert partition_count(spark.range(0, 10, 1, 4).coalesce(2)) == 2  # noqa: PLR2004



import pyarrow as pa
from pyiceberg.schema import Schema
from pyiceberg.types import IntegerType, NestedField, StringType, StructType

from .utils import create_sql_catalog  # noqa: TID252


def test_column_projection_subset_and_order(spark, tmp_path):
    catalog = create_sql_catalog(tmp_path)
    identifier = "default.test_projection_subset"
    schema = Schema(
        NestedField(1, "a", IntegerType(), required=False),
        NestedField(2, "b", StringType(), required=False),
        NestedField(3, "c", IntegerType(), required=False),
    )
    table = catalog.create_table(identifier=identifier, schema=schema)
    try:
        table.append(
            pa.table(
                {
                    "a": pa.array([1, 2], type=pa.int32()),
                    "b": pa.array(["x", "y"], type=pa.string()),
                    "c": pa.array([3, 4], type=pa.int32()),
                }
            )
        )
        path = table.location()

        df = spark.read.format("iceberg").load(path).select("c", "a")
        rows = sorted([(r[0], r[1]) for r in df.collect()])
        assert rows == [(3, 1), (4, 2)]
    finally:
        catalog.drop_table(identifier)


def test_nested_struct_projection_and_nulls(spark, tmp_path):
    catalog = create_sql_catalog(tmp_path)
    identifier = "default.test_projection_nested"
    inner = StructType(
        NestedField(10, "x", IntegerType(), required=False), NestedField(11, "y", StringType(), required=False)
    )
    schema = Schema(
        NestedField(1, "id", IntegerType(), required=False),
        NestedField(2, "s", inner, required=False),
    )
    table = catalog.create_table(identifier=identifier, schema=schema)
    try:
        struct_type = pa.struct([("x", pa.int32()), ("y", pa.string())])
        t1 = pa.Table.from_arrays(
            [
                pa.array([1], type=pa.int32()),
                pa.array([None], type=struct_type),
            ],
            names=["id", "s"],
        )
        t2 = pa.Table.from_arrays(
            [
                pa.array([2], type=pa.int32()),
                pa.array([{"x": 7, "y": "z"}], type=struct_type),
            ],
            names=["id", "s"],
        )
        table.append(t1)
        table.append(t2)
        path = table.location()

        df = spark.read.format("iceberg").load(path).select("id", "s.x")
        result = sorted([(r[0], r[1]) for r in df.collect()])
        assert result == [(1, None), (2, 7)]
    finally:
        catalog.drop_table(identifier)



import os

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from pyspark.sql.types import Row

from ..utils import assert_file_count_in_partitions, get_partition_structure  # noqa: TID252


class TestDeltaPartitioning:
    """Delta Lake partitioning functionality tests"""

    @pytest.fixture(scope="class")
    def delta_test_data(self):
        """Test data"""
        return [
            Row(id=10, event="A", score=0.98),
            Row(id=11, event="B", score=0.54),
            Row(id=12, event="A", score=0.76),
        ]

    def test_delta_partitioning_by_single_column(self, spark, tmp_path):
        """Test Delta Lake partitioning functionality"""
        delta_path = tmp_path / "delta_table"
        delta_table_path = f"{delta_path}"
        # Create partition data
        partition_data = [
            Row(id=1, event="A", year=2025, score=0.8),
            Row(id=2, event="B", year=2025, score=0.9),
            Row(id=3, event="A", year=2026, score=0.7),
            Row(id=4, event="B", year=2026, score=0.6),
        ]
        df = spark.createDataFrame(partition_data)

        # Write partitioned table
        df.write.format("delta").mode("overwrite").partitionBy("year").save(str(delta_path))

        # Physical artifact assertion: verify partition directory structure
        partition_dirs = get_partition_structure(str(delta_path))
        assert partition_dirs == {"year=2025", "year=2026"}, f"Expected year partitions, got {partition_dirs}"

        # Read entire table
        result_df = spark.read.format("delta").load(delta_table_path).sort("id")

        expected_data = pd.DataFrame(
            {
                "id": [1, 2, 3, 4],
                "event": ["A", "B", "A", "B"],
                "score": [0.8, 0.9, 0.7, 0.6],
                "year": [2025, 2025, 2026, 2026],
            }
        ).astype({"id": "int32", "event": "string", "score": "float64", "year": "int32"})

        assert_frame_equal(
            result_df.toPandas(), expected_data.sort_values("id").reset_index(drop=True), check_dtype=False
        )

        filtered_df = spark.read.format("delta").load(delta_table_path).filter("year = 2025").sort("id")
        expected = 2025
        expected_filtered = expected_data[expected_data["year"] == expected].sort_values("id").reset_index(drop=True)

        assert_frame_equal(filtered_df.toPandas(), expected_filtered, check_dtype=False)
        assert filtered_df.count() == 2, "Partition pruning should return exactly 2 records for year=2025"  # noqa: PLR2004

        filtered_df_ne = spark.read.format("delta").load(delta_table_path).filter("year != 2025")
        assert filtered_df_ne.count() == 2, "NOT EQUAL filter should return 2 records for year!=2025"  # noqa: PLR2004

        filtered_df_gt = spark.read.format("delta").load(delta_table_path).filter("year > 2025")
        assert filtered_df_gt.count() == 2, "GREATER THAN filter should return 2 records for year>2025"  # noqa: PLR2004

    def test_delta_partitioning_creates_correct_directory_structure(self, spark, delta_test_data, tmp_path):
        delta_path = tmp_path / "partitioned_delta_table"
        delta_table_path = f"{delta_path}"

        df = spark.createDataFrame(delta_test_data)

        df.write.format("delta").mode("overwrite").partitionBy("id").save(str(delta_path))

        result_df = spark.read.format("delta").load(delta_table_path).sort("id")

        expected_data = pd.DataFrame(
            {"event": ["A", "B", "A"], "score": [0.98, 0.54, 0.76], "id": [10, 11, 12]}
        ).astype({"event": "string", "score": "float64", "id": "int32"})

        result_pandas = result_df.toPandas()
        result_pandas = result_pandas.sort_values("id").reset_index(drop=True)
        expected_data = expected_data.sort_values("id").reset_index(drop=True)

        assert_frame_equal(result_pandas, expected_data, check_dtype=False)

        partition_dirs = []
        for item in os.listdir(delta_path):
            item_path = delta_path / item
            if os.path.isdir(item_path) and item.startswith("id="):
                partition_dirs.append(item)

        expected_partitions = {"id=10", "id=11", "id=12"}
        actual_partitions = set(partition_dirs)
        assert actual_partitions == expected_partitions, f"Expected {expected_partitions}, got {actual_partitions}"

        assert_file_count_in_partitions(str(delta_path), expected_files_per_partition=1)

    def test_delta_partitioning_by_multiple_columns(self, spark, tmp_path):
        """Test multi-column partitioning behavior."""
        delta_path = tmp_path / "multi_partitioned_delta_table"

        multi_partition_data = [
            Row(id=1, region=1, category=1, value=100),
            Row(id=2, region=1, category=2, value=200),
            Row(id=3, region=2, category=1, value=300),
            Row(id=4, region=2, category=2, value=400),
        ]

        df = spark.createDataFrame(multi_partition_data)

        df.write.format("delta").mode("overwrite").partitionBy("region", "category").save(str(delta_path))

        result_df = spark.read.format("delta").load(f"{delta_path}").sort("id")

        expected_data = pd.DataFrame(
            {"id": [1, 2, 3, 4], "value": [100, 200, 300, 400], "region": [1, 1, 2, 2], "category": [1, 2, 1, 2]}
        ).astype({"id": "int32", "value": "int32", "region": "int32", "category": "int32"})

        result_pandas = result_df.toPandas().sort_values("id").reset_index(drop=True)
        expected_data = expected_data.sort_values("id").reset_index(drop=True)

        assert_frame_equal(result_pandas, expected_data, check_dtype=False)

        expected_partition_structure = {
            "region=1/category=1",
            "region=1/category=2",
            "region=2/category=1",
            "region=2/category=2",
        }

        actual_partitions = set()
        for region_dir in os.listdir(delta_path):
            if region_dir.startswith("region="):
                region_path = delta_path / region_dir
                if os.path.isdir(region_path):
                    for category_dir in os.listdir(region_path):
                        if category_dir.startswith("category="):
                            actual_partitions.add(f"{region_dir}/{category_dir}")

        assert (
            actual_partitions == expected_partition_structure
        ), f"Expected {expected_partition_structure}, got {actual_partitions}"

        df_region1 = spark.read.format("delta").load(f"{delta_path}").filter("region = 1")
        assert df_region1.count() == 2, "Region 1 should have 2 records"  # noqa: PLR2004

        df_region2_cat2 = spark.read.format("delta").load(f"{delta_path}").filter("region = 2 AND category = 2")
        assert df_region2_cat2.count() == 1, "Region 2, Category 2 should have 1 record"

        df_region_ge2 = spark.read.format("delta").load(f"{delta_path}").filter("region >= 2")
        assert df_region_ge2.count() == 2, "Region >= 2 should have 2 records"  # noqa: PLR2004



import contextlib
import platform
import socket
import threading
import time

import pytest

from pysail.tests.spark.utils import is_jvm_spark

if is_jvm_spark():
    pytest.skip("JVM spark does not support the socket format in SQL", allow_module_level=True)


class SocketServer:
    def __init__(self, host="localhost", port=0, data=()):
        self.host = host
        self.port = port
        self.data = list(data)
        self._socket = None
        self._thread = None
        self._running = False

    def start(self):
        if self._running:
            return
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._socket.bind((self.host, self.port))
        (_, self.port) = self._socket.getsockname()
        self._socket.listen(1)

        self._running = True
        self._thread = threading.Thread(target=self._run_server)
        self._thread.daemon = True
        self._thread.start()

    def stop(self):
        self._running = False
        if self._socket is not None:
            # We need to explicitly shut down the socket to unblock `accept()`,
            # otherwise the thread may hang, and we will wait indefinitely in `join()`.
            with contextlib.suppress(OSError):
                self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
            self._socket = None
        if self._thread is not None:
            self._thread.join()
            self._thread = None

    def _run_server(self):
        while self._running:
            try:
                client_socket, _ = self._socket.accept()
                client_socket.settimeout(1)
                try:
                    for line in self.data:
                        if not self._running:
                            break
                        client_socket.send(f"{line}\n".encode())
                finally:
                    client_socket.close()
            except OSError:
                continue

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def probe(self, attempt=3, interval=0.1) -> bool:
        for _ in range(attempt):
            try:
                with socket.create_connection((self.host, self.port)):
                    return True
            except (ConnectionRefusedError, OSError):
                time.sleep(interval)
        return False


@pytest.fixture(scope="module", autouse=True)
def tables(spark):
    with SocketServer(data=["hello", "world"]) as server:
        if not server.probe():
            msg = "the socket server is not ready"
            raise RuntimeError(msg)

        spark.sql(f"""
            CREATE TABLE t1
            USING socket
            OPTIONS (
                host '{server.host}',
                port '{server.port}'
            )
        """)
        spark.sql("""
            CREATE TABLE t2_invalid_port
            USING socket
            OPTIONS (
                host 'localhost',
                port '0'
            )
        """)
        spark.sql("""
            CREATE TABLE t3_connection_failure
            USING socket
            OPTIONS (
                host 'localhost',
                port '65535'
            )
        """)
        yield
        spark.sql("DROP TABLE t1")
        spark.sql("DROP TABLE t2_invalid_port")
        spark.sql("DROP TABLE t3_connection_failure")


def test_socket_basic(spark):
    actual = spark.sql("SELECT * FROM t1 LIMIT 2").collect()
    assert actual == [("hello",), ("world",)]


def test_socket_with_projection(spark):
    actual = spark.sql("SELECT value, value FROM t1 LIMIT 1").collect()
    assert actual == [("hello", "hello")]


def test_socket_with_filtering(spark):
    actual = spark.sql("SELECT * FROM t1 WHERE value LIKE '%w%' LIMIT 1").collect()
    assert actual == [("world",)]


def test_socket_invalid_port(spark):
    with pytest.raises(Exception, match="port"):
        spark.sql("SELECT * FROM t2_invalid_port LIMIT 1").collect()


@pytest.mark.skipif(platform.system() == "Windows", reason="not working on Windows")
def test_socket_connection_failure(spark):
    with pytest.raises(Exception, match="refused"):
        spark.sql("SELECT * FROM t3_connection_failure LIMIT 1").collect()
